#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Plot doalamo output including confidence intervals if they are calculated.
"""

from idaes.surrogate.alamopy import almerror

def almplot(res, show=True):
    try:
        import matplotlib.pyplot as plt
        import numpy as np
    except Exception:
        raise almerror.AlamoInputError(
            "Cannot plot, possibly missing matplotlib package"
        )

    model = res['model']

    if type(model) is dict:
        for key in list(res['model'].keys()):
            model = res['model'][key].replace(' - ', ' + -')

            model = model.split('=')[1]
            model = model.split(' + ')
            if model[0] == ' ':  # if there are more than one terms, the first split is ' '
                model = model[1:]

            ndp = 100
            t = np.linspace(0.08, 1.7, ndp)
            out = np.zeros([3, ndp])
            clo = np.zeros(ndp)
            chi = np.zeros(ndp)
            coeff = np.zeros(ndp)

            for i in range(len(model)):
                thisterm = model[i].split(' * ')
                if thisterm[0].isdigit():
                    coeff[i] = float(model[i].split(' * ')[0])
                else:
                    coeff[i] = 1
                if 'conf_inv' in res.keys():
                    if type(res['conf_inv']) is dict:
                        # key = list(res['conf_inv'].keys())[0]
                        clo[i] = coeff[i] - float(res['conf_inv'][key][i].split('+/-')[1])
                        chi[i] = coeff[i] + float(res['conf_inv'][key][i].split('+/-')[1])
                    else:
                        clo[i] = coeff[i] - float(res['conf_inv'][i].split('+/-')[1])
                        chi[i] = coeff[i] + float(res['conf_inv'][i].split('+/-')[1])

            for i in range(ndp):
                out[0, i] = float(coeff[0]) * t[i] ** 1.2 \
                            - float(coeff[1]) * t[i] ** 2 - float(coeff[2])
                if 'conf_inv' in res.keys():  # If confidence intervals exist
                    out[1, i] = clo[0] * t[i] ** 1.2 - chi[1] * t[i] ** 2 - chi[2]
                    out[2, i] = chi[0] * t[i] ** 1.2 - clo[1] * t[i] ** 2 - clo[2]

            plt.plot(t, out[0], 'b-')
            if 'conf_inv' in res.keys():
                plt.plot(t, out[1], 'r--', t, out[2], 'r--')

            plt.xlabel(res['xlabels'][0])
            plt.ylabel(key)

            if show:
                plt.show()
    else:
        model = res['model'].replace(' - ', ' + -')

        model = model.split('=')[1]
        model = model.split(' + ')
        if model[0] == ' ':  # if there are more than one terms, the first split is ' '
            model = model[1:]
        ndp = 100
        t = np.linspace(0.08, 1.7, ndp)
        out = np.zeros([3, ndp])
        clo = np.zeros(ndp)
        chi = np.zeros(ndp)
        coeff = np.zeros(ndp)

        for i in range(len(model)):
            thisterm = model[i].split(' * ')
            if thisterm[0].isdigit():
                coeff[i] = float(model[i].split(' * ')[0])
            else:
                coeff[i] = 1
            if 'conf_inv' in res.keys():
                if type(res['conf_inv']) is dict:
                    firstKey = list(res['conf_inv'].keys())[0]
                    clo[i] = coeff[i] - float(res['conf_inv'][firstKey][i].split('+/-')[1])
                    chi[i] = coeff[i] + float(res['conf_inv'][firstKey][i].split('+/-')[1])
                else:
                    clo[i] = coeff[i] - float(res['conf_inv'][i].split('+/-')[1])
                    chi[i] = coeff[i] + float(res['conf_inv'][i].split('+/-')[1])

        for i in range(ndp):
            out[0, i] = float(coeff[0]) * t[i]**1.2 \
                - float(coeff[1]) * t[i]**2 - float(coeff[2])
            if 'conf_inv' in res.keys():  # If confidence intervals exist
                out[1, i] = clo[0] * t[i]**1.2 - chi[1] * t[i]**2 - chi[2]
                out[2, i] = chi[0] * t[i]**1.2 - clo[1] * t[i]**2 - clo[2]

        plt.plot(t, out[0], 'b-')
        if 'conf_inv' in res.keys():
            plt.plot(t, out[1], 'r--', t, out[2], 'r--')

        plt.xlabel(res['xlabels'][0])
        plt.ylabel(res['zlabels'][0])
        if show:
            plt.show()


#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Hydrocarbon processing vapor phase package using Peng-Robinson EoS.

Example property package using the Generic Property Package Framework.
This example shows how to set up a property package to do hydrocarbon
processing phase equilibrium in the generic framework using Peng-Robinson
equation along with methods drawn from the pre-built IDAES property libraries.

The example includes the dictionary named configuration contains parameters
for calculating vapor phase properties for hydrocarbon processing.
"""

# Import Python libraries
import logging

# Import Pyomo units
from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component
from idaes.core.phases import PhaseType as PT
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.eos.ceos import Cubic, CubicType
from idaes.generic_models.properties.core.phase_equil.forms import log_fugacity

from idaes.generic_models.properties.core.pure import Perrys
from idaes.generic_models.properties.core.pure import RPP4
from idaes.generic_models.properties.core.pure import RPP5

# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for a Prng Robinson alkene system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [2] Perry's Chemical Engineers' Handbook 7th Ed.
#     Converted to J/mol.K, mol/m^3
# [3] Engineering Toolbox, https://www.engineeringtoolbox.com
#     Retrieved 15th september, 2020
# [4] The Properties of Gases and Liquids (2001)
#     5th edition, Chemical Engineering Series - Robert C. Reid

configuration_vap = {
    # Specifying components
    "components": {
        "hydrogen": {"type": Component,
                    "elemental_composition": {'H': 2, 'C': 0},
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "valid_phase_types": PT.vaporPhase,
                    "parameter_data": {
                        "mw": (2.016E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (12.9e5, pyunits.Pa),  # [1]
                        "temperature_crit": (33.2, pyunits.K),  # [1]
                        "omega": -0.218,
                        "cp_mol_ig_comp_coeff": {
                            "A": (2.714e1, pyunits.J/pyunits.mol/pyunits.K),  # [1]
                            "B": (9.274e-3, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-1.981e-5, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (7.645e-9, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "entr_mol_form_vap_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            0.0, pyunits.J/pyunits.mol)}},

        "methane": {"type": Component,
                    "elemental_composition": {'H': 4, 'C': 1},
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "valid_phase_types": PT.vaporPhase,
                    "parameter_data": {
                        "mw": (16.043E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (46e5, pyunits.Pa),  # [1]
                        "temperature_crit": (190.4, pyunits.K),  # [1]
                        "omega": 0.011,
                        "cp_mol_ig_comp_coeff": {
                            "A": (1.925e1, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (5.213e-2, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (1.197e-5, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (-1.132e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "entr_mol_form_vap_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -74600, pyunits.J/pyunits.mol)}},

        "ethane": {"type": Component,
                    "elemental_composition": {'H': 6, 'C': 2},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (30.070E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (48.8e5, pyunits.Pa),  # [1]
                        "temperature_crit": (305.4, pyunits.K),  # [1]
                        "omega": 0.099,
                        "dens_mol_liq_comp_coeff": {
                            "1": (1.9122, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.27937, None),
                            "3": (305.32, pyunits.K),
                            "4": (0.29187, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (5.409e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (1.781e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-6.938e-5, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (8.713e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (90.9, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -83800, pyunits.J/pyunits.mol),  # [3]
                        "pressure_sat_comp_coeff": {"A": (3.95405, None),  # [4]
                                                    "B": (663.720, pyunits.K),
                                                    "C": (256.681, pyunits.K)}}},

        "propane": {"type": Component,
                    "elemental_composition": {'H': 8, 'C': 3},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (44.094E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (42.5e5, pyunits.Pa),  # [1]
                        "temperature_crit": (369.8, pyunits.K),  # [1]
                        "omega": 0.144,
                        "dens_mol_liq_comp_coeff": {
                            "1": (1.3757, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.274253, None),
                            "3": (369.83, pyunits.K),
                            "4": (0.29147, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (-4.224e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (3.063e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-1.586e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (3.215e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (112.71, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0.0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -104700, pyunits.J/pyunits.mol),  # [3]
                        "pressure_sat_comp_coeff": {"A": (3.92828, None),  # [4]
                                                    "B": (803.9970, pyunits.K),
                                                    "C": (247.040, pyunits.K)}}},

        "nbutane": {"type": Component,
                    "elemental_composition": {'H': 10, 'C': 4},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (58.124E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (38.0e5, pyunits.Pa),  # [1]
                        "temperature_crit": (425.2, pyunits.K),  # [1]
                        "omega": 0.199,
                        "dens_mol_liq_comp_coeff": {
                            "1": (1.0677, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.27188, None),
                            "3": (425.12, pyunits.K),
                            "4": (0.28688, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (9.487e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (3.313e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-1.108e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (-2.822e-9, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (132.42, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -125600, pyunits.J/pyunits.mol),  # [3]
                        "pressure_sat_comp_coeff": {"A": (3.93266, None),  # [4]
                                                    "B": (935.7730, pyunits.K),
                                                    "C": (238.789, pyunits.K)}}},

        "ibutane": {"type": Component,
                    "elemental_composition": {'H': 10, 'C': 4},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (58.124E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (36.5e5, pyunits.Pa),  # [1]
                        "temperature_crit": (408.2, pyunits.K),  # [1]
                        "omega": 0.183,
                        "dens_mol_liq_comp_coeff": {
                            "1": (1.0463, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.27294, None),
                            "3": (408.14, pyunits.K),
                            "4": (0.27301, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (-1.390e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (3.847e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-1.846e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (2.895e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (1.7237e5, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (-1.7839e3, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (1.4759e1, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (-4.7909e-2, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (5.8050e-5, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -135600, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            310.1, pyunits.J/pyunits.mol/pyunits.K),  # [3]
                        "pressure_sat_comp_coeff": {"A": (4.00272, None),  # [4]
                                                    "B": (947.5400, pyunits.K),
                                                    "C": (248.870, pyunits.K)}}},

        "ethylene": {"type": Component,
                    "elemental_composition": {'H': 4, 'C': 2},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (28.054E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (50.5e5, pyunits.Pa),  # [1]
                        "temperature_crit": (282.4, pyunits.K),  # [1]
                        "omega": 0.089,
                        "dens_mol_liq_comp_coeff": {
                            "1": (2.0961, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.27657, None),
                            "3": (282.34, pyunits.K),
                            "4": (0.29147, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (3.806e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (1.566e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-8.348e-5, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (1.755e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (2.4739e5, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (-4.4280e3, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (4.0936e1, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (-1.6970e-1, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            52283.264, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            219.5, pyunits.J/pyunits.mol/pyunits.K),  # [3]
                        "pressure_sat_comp_coeff": {"A": (3.67374, None),  # [4]
                                                    "B": (528.6700, pyunits.K),
                                                    "C": (228.790, pyunits.K)}}},

        "propene": {"type": Component,
                    "elemental_composition": {'H': 6, 'C': 3},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (42.081E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (46.2e5, pyunits.Pa),  # [1]
                        "temperature_crit": (365.0, pyunits.K),  # [1]
                        "omega": 0.144,
                        "dens_mol_liq_comp_coeff": {
                            "1": (1.4094, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.26465, None),
                            "3": (365.57, pyunits.K),
                            "4": (0.295, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (3.710e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (2.345e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-1.160e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (2.205e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (7.1720e5, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (-3.8632e2, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (1.2348e0, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            20413.736, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            266.9, pyunits.J/pyunits.mol/pyunits.K),  # [3]
                        "pressure_sat_comp_coeff": {"A": (3.95606, None),  # [4]
                                                    "B": (789.6200, pyunits.K),
                                                    "C": (247.580, pyunits.K)}}},

        "butene": {"type": Component,
                    "elemental_composition": {'H': 8, 'C': 4},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (56.104E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (40.2e5, pyunits.Pa),  # [1]
                        "temperature_crit": (419.3, pyunits.K),  # [1]
                        "omega": 0.191,
                        "dens_mol_liq_comp_coeff": {
                            "1": (1.0972, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.2649, None),
                            "3": (419.95, pyunits.K),
                            "4": (0.29043, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (-2.994e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (3.532e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-1.990e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (4.463e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (1.3589e5, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (-4.7739e2, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (2.1835e0, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (-2.2230e-3, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            1171.52, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            305.6, pyunits.J/pyunits.mol/pyunits.K),  # [3]
                        "pressure_sat_comp_coeff": {"A": (3.96640, None),  # [4]
                                                    "B": (927.2100, pyunits.K),
                                                    "C": (238.630, pyunits.K)}}},

        "pentene": {"type": Component,
                    "elemental_composition": {'H': 10, 'C': 5},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (70.135E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (40.5e5, pyunits.Pa),  # [1]
                        "temperature_crit": (464.7, pyunits.K),  # [1]
                        "omega": 0.233,  # [1]
                        "dens_mol_liq_comp_coeff": {
                            "1": (0.9038, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.26648, None),
                            "3": (464.78, pyunits.K),
                            "4": (0.2905, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (-1.340e-1, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (4.329e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-2.317e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (4.681e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (1.5467e5, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (-4.2600e2, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (1.964, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (-1.8034e-3, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -20920.00, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            347.82, pyunits.J/pyunits.mol/pyunits.K),  # [3]
                        "pressure_sat_comp_coeff": {"A": (3.96914, None),  # [4]
                                                    "B": (1044.010, pyunits.K),
                                                    "C": (233.450, pyunits.K)}}},

        "hexene": {"type": Component,
                    "elemental_composition": {'H': 12, 'C': 6},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (84.162E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (31.7e5, pyunits.Pa),  # [1]
                        "temperature_crit": (504.0, pyunits.K),  # [1]
                        "omega": 0.285,  # [1]
                        "dens_mol_liq_comp_coeff": {
                            "1": (0.7389, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.26147, None),
                            "3": (504.03, pyunits.K),
                            "4": (0.2902, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (-1.746e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (5.309e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-2.903e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (6.054e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (1.9263e5, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (-5.7116e2, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (2.4004e0, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (-1.9758e-3, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -41672.64, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            385.97, pyunits.J/pyunits.mol/pyunits.K),  # [3]
                        "pressure_sat_comp_coeff": {"A": (3.98260, None),  # [4]
                                                    "B": (1148.620, pyunits.K),
                                                    "C": (225.340, pyunits.K)}}},

        "heptene": {"type": Component,
                    "elemental_composition": {'H': 14, 'C': 7},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (98.189E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (25.4e5, pyunits.Pa),  # [1]
                        "temperature_crit": (537.2, pyunits.K),  # [1]
                        "omega": 0.358,  # [1]
                        "dens_mol_liq_comp_coeff": {
                            "1": (0.63734, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.26319, None),
                            "3": (537.29, pyunits.K),
                            "4": (0.27375, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (-3.303e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (6.297e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-3.512e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (7.607E-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (1.8997e5, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (-1.5670e2, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (3.4300e-1, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (1.5222e-3, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -62299.76, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            424, pyunits.J/pyunits.mol/pyunits.K),  # [3]
                        "pressure_sat_comp_coeff": {"A": (4.02677, None),  # [4]
                                                    "B": (1258.340, pyunits.K),
                                                    "C": (219.300, pyunits.K)}}},

        "octene": {"type": Component,
                    "elemental_composition": {'H': 16, 'C': 8},
                    "dens_mol_liq_comp": Perrys,
                    "enth_mol_liq_comp": Perrys,
                    "entr_mol_liq_comp": Perrys,
                    "enth_mol_ig_comp": RPP4,
                    "entr_mol_ig_comp": RPP4,
                    "pressure_sat_comp": RPP5,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (112.216E-3, pyunits.kg/pyunits.mol),  # [1]
                        "pressure_crit": (26.2e5, pyunits.Pa),  # [1]
                        "temperature_crit": (566.6, pyunits.K),  # [1]
                        "omega": 0.386,  # [1]
                        "dens_mol_liq_comp_coeff": {
                            "1": (0.5871, pyunits.kmol*pyunits.m**-3),  # [2] pg. 2-98
                            "2": (0.27005, None),
                            "3": (566.65, pyunits.K),
                            "4": (0.27187, None)},
                        "cp_mol_ig_comp_coeff": {
                            "A": (-4.099e0, pyunits.J/pyunits.mol/pyunits.K),
                            "B": (7.239e-1, pyunits.J/pyunits.mol/pyunits.K**2),
                            "C": (-4.036e-4, pyunits.J/pyunits.mol/pyunits.K**3),
                            "D": (8.675e-8, pyunits.J/pyunits.mol/pyunits.K**4)},
                        "cp_mol_liq_comp_coeff": {
                            "1": (3.7930e5, pyunits.J*pyunits.kmol**-1*pyunits.K**-1),  # [2]
                            "2": (-2.1175e3, pyunits.J*pyunits.kmol**-1*pyunits.K**-2),
                            "3": (8.2362e0, pyunits.J*pyunits.kmol**-1*pyunits.K**-3),
                            "4": (-9.0093e-3, pyunits.J*pyunits.kmol**-1*pyunits.K**-4),
                            "5": (0, pyunits.J*pyunits.kmol**-1*pyunits.K**-5)},
                        "enth_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_liq_comp_ref": (
                            0, pyunits.J/pyunits.mol/pyunits.K),
                        "enth_mol_form_vap_comp_ref": (
                            -82926.88, pyunits.J/pyunits.mol),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            454.36, pyunits.J/pyunits.mol/pyunits.K),  # [3]
                        "pressure_sat_comp_coeff": {"A": (4.05985, None),  # [4]
                                                    "B": (1355.460, pyunits.K),
                                                    "C": (213.050, pyunits.K)}}}},

    # Specifying phases
    "phases":  {'Vap': {"type": VaporPhase,
                        "equation_of_state": Cubic,
                        "equation_of_state_options": {
                            "type": CubicType.PR}}},

    # Set base units of measurement
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},

    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {"flow_mol": (0, 100, 1000, pyunits.mol/pyunits.s),
                     "temperature": (273.15, 300, 1500, pyunits.K),
                     "pressure": (5e4, 1e5, 1e7, pyunits.Pa)},
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),

    # Defining phase equilibria
    "parameter_data": {"PR_kappa": {("hydrogen", "hydrogen"): 0.000,
                                ("hydrogen", "methane"): 0.000,
                                ("hydrogen", "ethane"): 0.000,
                                ("hydrogen", "propane"): 0.000,
                                ("hydrogen", "nbutane"): 0.000,
                                ("hydrogen", "ibutane"): 0.000,
                                ("hydrogen", "ethylene"): 0.000,
                                ("hydrogen", "propene"): 0.000,
                                ("hydrogen", "butene"): 0.000,
                                ("hydrogen", "pentene"): 0.000,
                                ("hydrogen", "hexene"): 0.000,
                                ("hydrogen", "heptene"): 0.000,
                                ("hydrogen", "octene"): 0.000,
                                ("methane", "hydrogen"): 0.000,
                                ("methane", "methane"): 0.000,
                                ("methane", "ethane"): 0.000,
                                ("methane", "propane"): 0.000,
                                ("methane", "nbutane"): 0.000,
                                ("methane", "ibutane"): 0.000,
                                ("methane", "ethylene"): 0.000,
                                ("methane", "propene"): 0.000,
                                ("methane", "butene"): 0.000,
                                ("methane", "pentene"): 0.000,
                                ("methane", "hexene"): 0.000,
                                ("methane", "heptene"): 0.000,
                                ("methane", "octene"): 0.000,
                                ("ethane", "hydrogen"): 0.000,
                                ("ethane", "methane"): 0.000,
                                ("ethane", "ethane"): 0.000,
                                ("ethane", "propane"): 0.000,
                                ("ethane", "nbutane"): 0.000,
                                ("ethane", "ibutane"): 0.000,
                                ("ethane", "ethylene"): 0.000,
                                ("ethane", "propene"): 0.000,
                                ("ethane", "butene"): 0.000,
                                ("ethane", "pentene"): 0.000,
                                ("ethane", "hexene"): 0.000,
                                ("ethane", "heptene"): 0.000,
                                ("ethane", "octene"): 0.000,
                                ("propane", "hydrogen"): 0.000,
                                ("propane", "methane"): 0.000,
                                ("propane", "ethane"): 0.000,
                                ("propane", "propane"): 0.000,
                                ("propane", "nbutane"): 0.000,
                                ("propane", "ibutane"): 0.000,
                                ("propane", "ethylene"): 0.000,
                                ("propane", "propene"): 0.000,
                                ("propane", "butene"): 0.000,
                                ("propane", "pentene"): 0.000,
                                ("propane", "hexene"): 0.000,
                                ("propane", "heptene"): 0.000,
                                ("propane", "octene"): 0.000,
                                ("nbutane", "hydrogen"): 0.000,
                                ("nbutane", "methane"): 0.000,
                                ("nbutane", "ethane"): 0.000,
                                ("nbutane", "propane"): 0.000,
                                ("nbutane", "nbutane"): 0.000,
                                ("nbutane", "ibutane"): 0.000,
                                ("nbutane", "ethylene"): 0.000,
                                ("nbutane", "propene"): 0.000,
                                ("nbutane", "butene"): 0.000,
                                ("nbutane", "pentene"): 0.000,
                                ("nbutane", "hexene"): 0.000,
                                ("nbutane", "heptene"): 0.000,
                                ("nbutane", "octene"): 0.000,
                                ("ibutane", "hydrogen"): 0.000,
                                ("ibutane", "methane"): 0.000,
                                ("ibutane", "ethane"): 0.000,
                                ("ibutane", "propane"): 0.000,
                                ("ibutane", "nbutane"): 0.000,
                                ("ibutane", "ibutane"): 0.000,
                                ("ibutane", "ethylene"): 0.000,
                                ("ibutane", "propene"): 0.000,
                                ("ibutane", "butene"): 0.000,
                                ("ibutane", "pentene"): 0.000,
                                ("ibutane", "hexene"): 0.000,
                                ("ibutane", "heptene"): 0.000,
                                ("ibutane", "octene"): 0.000,
                                ("ethylene", "hydrogen"): 0.000,
                                ("ethylene", "methane"): 0.000,
                                ("ethylene", "ethane"): 0.000,
                                ("ethylene", "propane"): 0.000,
                                ("ethylene", "nbutane"): 0.000,
                                ("ethylene", "ibutane"): 0.000,
                                ("ethylene", "ethylene"): 0.000,
                                ("ethylene", "propene"): 0.000,
                                ("ethylene", "butene"): 0.000,
                                ("ethylene", "pentene"): 0.000,
                                ("ethylene", "hexene"): 0.000,
                                ("ethylene", "heptene"): 0.000,
                                ("ethylene", "octene"): 0.000,
                                ("propene", "hydrogen"): 0.000,
                                ("propene", "methane"): 0.000,
                                ("propene", "ethane"): 0.000,
                                ("propene", "propane"): 0.000,
                                ("propene", "nbutane"): 0.000,
                                ("propene", "ibutane"): 0.000,
                                ("propene", "ethylene"): 0.000,
                                ("propene", "propene"): 0.000,
                                ("propene", "butene"): 0.000,
                                ("propene", "pentene"): 0.000,
                                ("propene", "hexene"): 0.000,
                                ("propene", "heptene"): 0.000,
                                ("propene", "octene"): 0.000,
                                ("butene", "hydrogen"): 0.000,
                                ("butene", "methane"): 0.000,
                                ("butene", "ethane"): 0.000,
                                ("butene", "propane"): 0.000,
                                ("butene", "nbutane"): 0.000,
                                ("butene", "ibutane"): 0.000,
                                ("butene", "ethylene"): 0.000,
                                ("butene", "propene"): 0.000,
                                ("butene", "butene"): 0.000,
                                ("butene", "pentene"): 0.000,
                                ("butene", "hexene"): 0.000,
                                ("butene", "heptene"): 0.000,
                                ("butene", "octene"): 0.000,
                                ("pentene", "hydrogen"): 0.000,
                                ("pentene", "methane"): 0.000,
                                ("pentene", "ethane"): 0.000,
                                ("pentene", "propane"): 0.000,
                                ("pentene", "nbutane"): 0.000,
                                ("pentene", "ibutane"): 0.000,
                                ("pentene", "ethylene"): 0.000,
                                ("pentene", "propene"): 0.000,
                                ("pentene", "butene"): 0.000,
                                ("pentene", "pentene"): 0.000,
                                ("pentene", "hexene"): 0.000,
                                ("pentene", "heptene"): 0.000,
                                ("pentene", "octene"): 0.000,
                                ("hexene", "hydrogen"): 0.000,
                                ("hexene", "methane"): 0.000,
                                ("hexene", "ethane"): 0.000,
                                ("hexene", "propane"): 0.000,
                                ("hexene", "nbutane"): 0.000,
                                ("hexene", "ibutane"): 0.000,
                                ("hexene", "ethylene"): 0.000,
                                ("hexene", "propene"): 0.000,
                                ("hexene", "butene"): 0.000,
                                ("hexene", "pentene"): 0.000,
                                ("hexene", "hexene"): 0.000,
                                ("hexene", "heptene"): 0.000,
                                ("hexene", "octene"): 0.000,
                                ("heptene", "hydrogen"): 0.000,
                                ("heptene", "methane"): 0.000,
                                ("heptene", "ethane"): 0.000,
                                ("heptene", "propane"): 0.000,
                                ("heptene", "nbutane"): 0.000,
                                ("heptene", "ibutane"): 0.000,
                                ("heptene", "ethylene"): 0.000,
                                ("heptene", "propene"): 0.000,
                                ("heptene", "butene"): 0.000,
                                ("heptene", "pentene"): 0.000,
                                ("heptene", "hexene"): 0.000,
                                ("heptene", "heptene"): 0.000,
                                ("heptene", "octene"): 0.000,
                                ("octene", "hydrogen"): 0.000,
                                ("octene", "methane"): 0.000,
                                ("octene", "ethane"): 0.000,
                                ("octene", "propane"): 0.000,
                                ("octene", "nbutane"): 0.000,
                                ("octene", "ibutane"): 0.000,
                                ("octene", "ethylene"): 0.000,
                                ("octene", "propene"): 0.000,
                                ("octene", "butene"): 0.000,
                                ("octene", "pentene"): 0.000,
                                ("octene", "hexene"): 0.000,
                                ("octene", "heptene"): 0.000,
                                ("octene", "octene"): 0.000}}}



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for rate forms
"""

import pytest

from pyomo.environ import \
    ConcreteModel, units as pyunits, value, SolverStatus, TerminationCondition

# Import IDAES cores
from idaes.generic_models.properties.core.generic.generic_property import \
    GenericParameterBlock
from idaes.generic_models.properties.core.generic.generic_reaction import \
    GenericReactionParameterBlock, ConcentrationForm
from idaes.generic_models.unit_models import EquilibriumReactor
from idaes.core import \
    Component, EnergyBalanceType, FlowsheetBlock, LiquidPhase, SolidPhase
from idaes.core.util import get_solver
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

from idaes.generic_models.properties.core.state_definitions import FTPx, FpcTP
from idaes.generic_models.properties.core.eos.ideal import Ideal
from idaes.generic_models.properties.core.reactions.equilibrium_forms import \
    solubility_product
from idaes.generic_models.properties.core.reactions.equilibrium_constant import \
    ConstantKeq


solver = get_solver()


def dummy_h(b, *args, **kwargs):
    return b.temperature


thermo_config = {
    # Specifying components
    "components": {
        'H2O': {'type': Component,
                'enth_mol_liq_comp': dummy_h},
        'Na+': {"type": Component,
                "enth_mol_liq_comp": dummy_h},
        'Cl-': {"type": Component,
                "enth_mol_liq_comp": dummy_h},
        'NaCl': {"type": Component,
                 "enth_mol_sol_comp": dummy_h}},
    # Specifying phases
    "phases":  {'Liq': {"type": LiquidPhase,
                        "equation_of_state": Ideal,
                        "component_list": ["H2O", "Na+", "Cl-"]},
                'Sol': {"type": SolidPhase,
                        "equation_of_state": Ideal,
                        "component_list": ["NaCl"]}},
    # Set base units of measurement
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    # Specifying state definition
    "state_definition": FpcTP,
    "state_bounds": {
        "flow_mol_phase_comp": (0, 100, 1000, pyunits.mol/pyunits.s),
        "temperature": (273.15, 300, 450, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa)},
    "pressure_ref": (1e5, pyunits.Pa),
    "temperature_ref": (300, pyunits.K)}

# Solubility of NaCl in H2O is 360 g/L, or 6.16 mol/L
# 1L of H2O is 55.56 mol, results in Ksp = 8.235e-3 on mole frac basis
rxn_config = {
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "equilibrium_reactions": {
        "S1": {"stoichiometry": {("Liq", "Na+"): 1,
                                 ("Liq", "Cl-"): 1,
                                 ("Sol", "NaCl"): -1},
               "equilibrium_constant": ConstantKeq,
               "equilibrium_form": solubility_product,
               "concentration_form": ConcentrationForm.moleFraction,
               "parameter_data": {
                   "k_eq_ref": (8.235e-3, None)}}}}


class TestSingleState(object):
    @pytest.fixture
    def model(self):
        m = ConcreteModel()

        # Add a test thermo package for validation
        m.pparams = GenericParameterBlock(default=thermo_config)
        m.rparams = GenericReactionParameterBlock(default={
            "property_package": m.pparams, **rxn_config})

        m.state = m.pparams.build_state_block(
            [0], default={"defined_state": False})

        m.rxn = m.rparams.build_reaction_block(
            [0], default={"state_block": m.state, "has_equilibrium": True})

        return m

    @pytest.mark.component
    def test_saturated(self, model):
        assert model.state[0].phase_component_set == [
            ("Liq", "H2O"), ("Liq", "Na+"), ("Liq", "Cl-"), ("Sol", "NaCl")]

        model.state[0].temperature.fix(298.15)
        model.state[0].pressure.fix(101325)

        # Solve for saturated states (i.e. solubility product applied) by
        # fixing the flowrate of solids to a positive value.
        # Then start with a large amount of Na+ and solve for the flowrate of
        # Cl- which satisfies the solubility product (ignore electroneutrality)
        model.state[0].flow_mol_phase_comp["Liq", "H2O"].fix(55.56)
        model.state[0].flow_mol_phase_comp["Liq", "Na+"].fix(20)
        model.state[0].flow_mol_phase_comp["Liq", "Cl-"].set_value(2.5)
        model.state[0].flow_mol_phase_comp["Sol", "NaCl"].fix(1)

        for i in range(20, 2, -1):
            model.state[0].flow_mol_phase_comp["Liq", "Na+"].fix(i)

            results = solver.solve(model)

            assert results.solver.termination_condition == \
                TerminationCondition.optimal
            assert results.solver.status == SolverStatus.ok

            assert pytest.approx(8.235e-3, abs=1e-8) == value(
                model.state[0].mole_frac_phase_comp["Liq", "Na+"] *
                model.state[0].mole_frac_phase_comp["Liq", "Cl-"])

    @pytest.mark.component
    def test_subsaturated(self, model):
        assert model.state[0].phase_component_set == [
            ("Liq", "H2O"), ("Liq", "Na+"), ("Liq", "Cl-"), ("Sol", "NaCl")]

        model.state[0].temperature.fix(298.15)
        model.state[0].pressure.fix(101325)

        # For subsaturated systems, fix the Na and Cl flows and check that no
        # solid is formed.
        model.state[0].flow_mol_phase_comp["Liq", "H2O"].fix(55.56)
        model.state[0].flow_mol_phase_comp["Liq", "Na+"].fix(0)
        model.state[0].flow_mol_phase_comp["Liq", "Cl-"].fix(0)
        model.state[0].flow_mol_phase_comp["Sol", "NaCl"].set_value(0)

        for i in range(11):
            for j in range(11):
                if i*j < 40:
                    # Zeroes result in evaluation errors, so use a small number
                    if i == 0:
                        i = 1e-8
                    if j == 0:
                        j = 1e-8

                    model.state[0].flow_mol_phase_comp["Liq", "Na+"].fix(i)
                    model.state[0].flow_mol_phase_comp["Liq", "Cl-"].fix(j)

                    results = solver.solve(model)

                    assert results.solver.termination_condition == \
                        TerminationCondition.optimal
                    assert results.solver.status == SolverStatus.ok

                    assert pytest.approx(0, abs=1e-5) == value(
                        model.state[0].flow_mol_phase_comp["Sol", "NaCl"])


class TestUnit(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        # Add a test thermo package for validation
        m.fs.pparams = GenericParameterBlock(default=thermo_config)
        m.fs.rparams = GenericReactionParameterBlock(default={
            "property_package": m.fs.pparams, **rxn_config})

        # Don't include energy balances, as the test doesn't have a proper
        # enthalpy model. Fix outlet T instead.
        m.fs.R101 = EquilibriumReactor(default={
            "property_package": m.fs.pparams,
            "reaction_package": m.fs.rparams,
            "has_equilibrium_reactions": True,
            "has_rate_reactions": False,
            "energy_balance_type": EnergyBalanceType.none})

        m.fs.R101.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(55.56)
        m.fs.R101.inlet.flow_mol_phase_comp[0, "Liq", "Na+"].fix(1e-8)
        m.fs.R101.inlet.flow_mol_phase_comp[0, "Liq", "Cl-"].fix(1e-8)
        m.fs.R101.inlet.flow_mol_phase_comp[0, "Sol", "NaCl"].fix(1e-8)
        m.fs.R101.inlet.temperature[0].fix(298.15)
        m.fs.R101.inlet.pressure[0].fix(101325)

        m.fs.R101.outlet.temperature[0].fix(298.15)

        return m

    @pytest.mark.integration
    def test_subsaturated(self, model):
        for i in range(0, 6):
            # Values of zero cause probelms, use a small number instead
            if i == 0:
                i == 1e-8

            model.fs.R101.inlet.flow_mol_phase_comp[0, "Sol", "NaCl"].fix(i)

            results = solver.solve(model)

            assert results.solver.termination_condition == \
                TerminationCondition.optimal
            assert results.solver.status == SolverStatus.ok

            assert pytest.approx(0, abs=1.1e-6) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Sol", "NaCl"])
            assert pytest.approx(i, rel=1e-6, abs=1e-6) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Na+"])
            assert pytest.approx(i, rel=1e-6, abs=1e-6) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Cl-"])

    @pytest.mark.integration
    def test_saturated(self, model):
        model.fs.R101.inlet.flow_mol_phase_comp[0, "Sol", "NaCl"].fix(6.16)

        results = solver.solve(model)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert pytest.approx(0, abs=1.1e-3) == value(
            model.fs.R101.outlet.flow_mol_phase_comp[0, "Sol", "NaCl"])
        assert pytest.approx(6.158968, rel=1e-5) == value(
            model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Na+"])
        assert pytest.approx(6.158968, rel=1e-5) == value(
            model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Cl-"])

    @pytest.mark.integration
    def test_supersaturated(self, model):
        for i in range(7, 21):
            model.fs.R101.inlet.flow_mol_phase_comp[0, "Sol", "NaCl"].fix(i)

            results = solver.solve(model)

            assert results.solver.termination_condition == \
                TerminationCondition.optimal
            assert results.solver.status == SolverStatus.ok

            assert pytest.approx(i, rel=1e-5) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Sol", "NaCl"] +
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Na+"])
            assert pytest.approx(6.159876, rel=1e-5) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Na+"])
            assert pytest.approx(6.159876, rel=1e-5) == value(
                model.fs.R101.outlet.flow_mol_phase_comp[0, "Liq", "Cl-"])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Simplified Flash Unit Model, only for IAPWS with mixed state.
Phase separator: inlet water/steam mixture is separated into liquid and vapor
streams.

Created: August 21 2020

Created on Thu Aug 18 12:59:50 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest
# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.generic_models.properties import iapws95

from idaes.power_generation.unit_models.helm.phase_separator import \
    HelmPhaseSeparator
from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture
def build_phase_separator():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmPhaseSeparator(default={'property_package':
                                            m.fs.properties})
    return m


@pytest.mark.unit
def test_basic_build(build_phase_separator):
    """Make a model and make sure it doesn't throw exception"""
    m = build_phase_separator
    assert degrees_of_freedom(m) == 3
    # Check unit config arguments
    assert len(m.fs.unit.config) == 4
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize(build_phase_separator):
    state_args_water_steam = {'flow_mol': 1.5e5,  # mol/s
                              'pressure': 1.2e7,  # Pa
                              'enth_mol': 28365.2608}  # j/mol
    initialization_tester(build_phase_separator, dof=3,
                          state_args_water_steam=state_args_water_steam)


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_wflash(build_phase_separator):
    m = build_phase_separator
    state_args_water_steam = {'flow_mol': 1.5e5,  # mol/s
                              'pressure': 1.2e7,  # Pa
                              'enth_mol': 28365.2608}  # j/mol
    m.fs.unit.initialize(state_args_water_steam=state_args_water_steam)
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    solver.solve(m, tee=True)
    Fin = pyo.value(m.fs.unit.mixed_state[0].flow_mol)
    FL = pyo.value(m.fs.unit.liq_state[0].flow_mol)
    FV = pyo.value(m.fs.unit.vap_state[0].flow_mol)
    assert degrees_of_freedom(m) == 0
    assert (pytest.approx((FL + FV), abs=1e-3) == Fin)

    assert (pytest.approx(pyo.value(m.fs.unit.mixed_state[0].
                                    flow_mol * m.fs.unit.mixed_state[0].
                                    vapor_frac), abs=1e-3)
            == pyo.value(m.fs.unit.vap_state[0].flow_mol))

    assert (pytest.approx(pyo.value(m.fs.unit.mixed_state[0].
                                    enth_mol_phase["Vap"]), abs=1e-3)
            == pyo.value(m.fs.unit.vap_state[0].enth_mol))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Library of common forms for phase equilibrium constraints
"""
from pyomo.environ import ConcreteModel, Var, units as pyunits

from idaes.generic_models.properties.core.generic.generic_property import \
    GenericParameterBlock
from idaes.generic_models.properties.core.state_definitions import FTPx

from idaes.generic_models.properties.core.phase_equil.forms import *
import pytest


# Dummy EoS to use for fugacity calls
class DummyEoS(object):
    def common(self):
        pass

    def build_parameters(b):
        pass

    def fug_phase_comp_eq(b, p, j, pp):
        return b.x[p, j]

    def log_fug_phase_comp_eq(b, p, j, pp):
        return 42*b.x[p, j]


@pytest.mark.unit
def test_fugacity():
    m = ConcreteModel()

    # Add a dummy var for use in constructing expressions
    m.x = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    m.mole_frac_phase_comp = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    # Create a dummy parameter block
    m.params = GenericParameterBlock(default={
        "components": {"H2O": {"parameter_data": {"temperature_crit": 647.3},
                               "phase_equilibrium_form": {
                                   ("Vap", "Liq"): fugacity}}},
        "phases": {"Liq": {"equation_of_state": DummyEoS},
                   "Vap": {"equation_of_state": DummyEoS}},
        "state_definition": FTPx,
        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K}})

    assert str(fugacity.return_expression(m, "Vap", "Liq", "H2O")) == str(
        m.x["Vap", "H2O"] == m.x["Liq", "H2O"])


@pytest.mark.unit
def test_log_fugacity():
    m = ConcreteModel()

    # Add a dummy var for use in constructing expressions
    m.x = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    m.mole_frac_phase_comp = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    # Create a dummy parameter block
    m.params = GenericParameterBlock(default={
        "components": {"H2O": {"parameter_data": {"temperature_crit": 647.3},
                               "phase_equilibrium_form": {
                                   ("Vap", "Liq"): log_fugacity}}},
        "phases": {"Liq": {"equation_of_state": DummyEoS},
                   "Vap": {"equation_of_state": DummyEoS}},
        "state_definition": FTPx,
        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K}})

    assert str(log_fugacity.return_expression(m, "Vap", "Liq", "H2O")) == str(
        42*m.x["Vap", "H2O"] == 42*m.x["Liq", "H2O"])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"Tests used to test the functionality of the test suite itself."

import pytest


# pytestmark = pytest.mark.skip("Tests for the test suite are always skipped under normal circumstances")


@pytest.mark.xfail
def test_function_without_any_required_mark():
    assert True


@pytest.mark.xfail
@pytest.mark.unit
@pytest.mark.component
def test_function_with_too_many_required_marks():
    assert True



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for generic property package core code

Author: Andrew Lee
"""
import pytest
from sys import modules

from pyomo.environ import (value, Block, ConcreteModel, Param,
                           Set, Var, units as pyunits)

from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterData,
        GenericStateBlock)
from idaes.generic_models.properties.core.generic.tests.dummy_eos import DummyEoS

from idaes.core import (declare_process_block_class, Component,
                        Phase, LiquidPhase, VaporPhase, MaterialFlowBasis)
from idaes.core.phases import PhaseType as PT
from idaes.core.util.exceptions import (ConfigurationError)
import idaes.logger as idaeslog


# -----------------------------------------------------------------------------
# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


def calculate_scaling_factors(b):
    b.scaling_check = True


# Dummy build_parameter methods for tests
def build_parameters(cobj, p):
    cobj.add_component("test_param_"+p, Var(initialize=42))


# Declare a base units dict to save code later
base_units = {"time": pyunits.s,
              "length": pyunits.m,
              "mass": pyunits.kg,
              "amount": pyunits.mol,
              "temperature": pyunits.K}


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True


class TestGenericParameterBlock(object):
    @pytest.mark.unit
    def test_build(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

        assert m.params.configured

        assert m.params.get_metadata().default_units == {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
            "current": None,
            "luminous intensity": None}

        assert isinstance(m.params.component_list, Set)
        assert len(m.params.component_list) == 3
        for j in m.params.component_list:
            assert j in ["a", "b", "c"]
            assert isinstance(m.params.get_component(j), Component)

        assert isinstance(m.params.phase_list, Set)
        assert len(m.params.phase_list) == 2
        for p in m.params.phase_list:
            assert p in ["p1", "p2"]
        assert isinstance(m.params.get_phase("p1"), LiquidPhase)
        assert isinstance(m.params.get_phase("p2"), Phase)
        assert m.params.p1.config.equation_of_state == DummyEoS
        assert m.params.p2.config.equation_of_state == DummyEoS

        assert isinstance(m.params._phase_component_set, Set)
        assert len(m.params._phase_component_set) == 5
        for i in m.params._phase_component_set:
            assert i in [("p1", "a"), ("p1", "b"),
                         ("p2", "a"), ("p2", "b"), ("p2", "c")]

        assert isinstance(m.params.pressure_ref, Param)
        assert m.params.pressure_ref.value == 1e5
        assert isinstance(m.params.temperature_ref, Param)
        assert m.params.temperature_ref.value == 300

        assert m.params.state_block_class is GenericStateBlock

        assert len(m.params.config.inherent_reactions) == 0
        assert m.params.config.reaction_basis == MaterialFlowBasis.molar

        assert not m.params.has_inherent_reactions

    @pytest.mark.unit
    def test_invalid_unit(self):
        m = ConcreteModel()

        with pytest.raises(
                ConfigurationError,
                match="params recieved unexpected units for quantity time: "
                "foo. Units must be instances of a Pyomo unit object."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"time": "foo",
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

    @pytest.mark.unit
    def test_missing_required_quantity(self):
        m = ConcreteModel()

        with pytest.raises(
                ConfigurationError,
                match="params units for quantity time were not assigned. "
                "Please make sure to provide units for all base units "
                "when configuring the property package."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})

    @pytest.mark.unit
    def test_no_components(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params was not provided with a components "
                           "argument."):
            m.params = DummyParameterBlock(default={
                "phases": {
                        "p1": {"equation_of_state": "foo"},
                        "p2": {"equation_of_state": "bar"}},
                "base_units": base_units})

    @pytest.mark.unit
    def test_no_phases(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params was not provided with a phases "
                           "argument."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "base_units": base_units})

    @pytest.mark.unit
    def test_invalid_component_in_phase_component_list(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params phase-component list for phase p1 "
                           "contained component d which is not in the master "
                           "component list"):
            m.params = DummyParameterBlock(default={
                    "components": {"a": {}, "b": {}, "c": {}},
                    "phases": {
                        "p1": {"type": LiquidPhase,
                               "component_list": ["a", "d"],
                               "equation_of_state": "foo"},
                        "p2": {"equation_of_state": "bar"}},
                    "base_units": base_units})

    @pytest.mark.unit
    def test_invalid_orphaned_component_in_phase_component_list(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params Component c does not appear to be "
                           "valid in any phase. Please check the component "
                           "lists defined for each phase, and be sure you do "
                           "not have generic Components in single-phase "
                           "aqueous systems."):
            m.params = DummyParameterBlock(default={
                    "components": {"a": {}, "b": {}, "c": {}},
                    "phases": {
                        "p1": {"type": LiquidPhase,
                               "component_list": ["a", "b"],
                               "equation_of_state": "foo"}},
                    "base_units": base_units})

    @pytest.mark.unit
    def test_invalid_component_in_phase_component_list_2(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params phase-component list for phase p1 "
                           "contained component a, however this component is "
                           "not valid for the given PhaseType"):
            m.params = DummyParameterBlock(default={
                    "components": {
                        "a": {"valid_phase_types": PT.solidPhase},
                        "b": {},
                        "c": {}},
                    "phases": {
                        "p1": {"type": LiquidPhase,
                               "component_list": ["a", "b"],
                               "equation_of_state": "foo"},
                        "p2": {"equation_of_state": "bar"}},
                    "base_units": base_units})

    @pytest.mark.unit
    def test_phase_component_set_from_valid_phases(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"valid_phase_types": PT.liquidPhase},
                    "b": {"valid_phase_types": PT.vaporPhase},
                    "c": {"valid_phase_types": [PT.liquidPhase,
                                                PT.vaporPhase]}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "equation_of_state": DummyEoS},
                    "p2": {"type": VaporPhase,
                           "equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

        assert len(m.params._phase_component_set) == 4
        for i in m.params._phase_component_set:
            assert i in [("p1", "a"), ("p1", "c"),
                         ("p2", "b"), ("p2", "c")]

    @pytest.mark.unit
    def test_no_state_definition(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params Generic Property Package was not "
                           "provided with a state_definition configuration "
                           "argument. Please fix your property parameter "
                           "definition to include this."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"}},
                "base_units": base_units})

    @pytest.mark.unit
    def test_no_pressure_ref(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params Generic Property Package was not "
                           "provided with a pressure_ref configuration "
                           "argument. Please fix your property parameter "
                           "definition to include this."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {}},
                "state_definition": "baz",
                "base_units": base_units})

    @pytest.mark.unit
    def test_convert_pressure_ref(self):
        m = ConcreteModel()

        # This will fail, but should set the reference pressure
        with pytest.raises(ConfigurationError):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {}},
                "state_definition": "baz",
                "pressure_ref": (1, pyunits.bar),
                "base_units": base_units})
        assert m.params.pressure_ref.value == 1e5

    @pytest.mark.unit
    def test_log_no_units_pressure_ref(self, caplog):
        m = ConcreteModel()

        caplog.set_level(
            idaeslog.DEBUG,
            logger=("idaes.generic_models.properties.core."
                    "generic.generic_property"))

        # This will fail, but should set the reference pressure
        with pytest.raises(ConfigurationError):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "base_units": base_units})
        assert m.params.pressure_ref.value == 1e5
        assert ('params no units provided for parameter pressure_ref - '
                'assuming default units' in caplog.text)

    @pytest.mark.unit
    def test_no_temperature_ref(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params Generic Property Package was not "
                           "provided with a temperature_ref configuration "
                           "argument. Please fix your property parameter "
                           "definition to include this."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "base_units": base_units})

    @pytest.mark.unit
    def test_log_no_units_temperature_ref(self, caplog):
        m = ConcreteModel()

        caplog.set_level(
            idaeslog.DEBUG,
            logger=("idaes.generic_models.properties.core."
                    "generic.generic_property"))

        # This will fail, but should set the reference pressure
        with pytest.raises(ConfigurationError):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})
        assert m.params.temperature_ref.value == 300
        assert ('params no units provided for parameter pressure_ref - '
                'assuming default units' in caplog.text)

    @pytest.mark.unit
    def test_convert_temperature_ref(self):
        m = ConcreteModel()

        # This will fail, but should set temerpature_ref
        with pytest.raises(ConfigurationError):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "temperature_ref": (540, pyunits.degR),
                "base_units": base_units})
        assert m.params.temperature_ref.value == 300

    @pytest.mark.unit
    def test_no_eos(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params phase p2 was not provided with an "
                           "equation_of_state configuration argument. Please "
                           "fix your property parameter definition to "
                           "include this."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

    @pytest.mark.unit
    def test_phases_in_equilibrium(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(default={
            "components": {
                "a": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                "b": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                "c": {"phase_equilibrium_form": {("p1", "p2"): "foo"}}},
            "phases": {
                "p1": {"equation_of_state": DummyEoS},
                "p2": {"equation_of_state": DummyEoS}},
            "state_definition": modules[__name__],
            "pressure_ref": 1e5,
            "temperature_ref": 300,
            "phases_in_equilibrium": [("p1", "p2")],
            "phase_equilibrium_state": {("p1", "p2"): "whoop"},
            "base_units": base_units})

        assert isinstance(m.params.phase_equilibrium_idx, Set)
        assert len(m.params.phase_equilibrium_idx) == 3
        for i in m.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2", "PE3"]

        assert isinstance(m.params.phase_equilibrium_list, dict)
        assert m.params.phase_equilibrium_list == {
            "PE1": {"a": ("p1", "p2")},
            "PE2": {"b": ("p1", "p2")},
            "PE3": {"c": ("p1", "p2")}}

    @pytest.mark.unit
    def test_phases_in_equilibrium_no_form(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params Generic Property Package component a "
                                 "is in equilibrium but phase_equilibrium_form"
                                 " was not specified."):
            m.params = DummyParameterBlock(default={
                "components": {
                    "a": {},
                    "b": {},
                    "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "phases_in_equilibrium": [("p1", "p2")],
                "phase_equilibrium_state": {("p1", "p2"): "whoop"},
                "base_units": base_units})

    @pytest.mark.unit
    def test_phases_in_equilibrium_missing_pair_form(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params Generic Property Package component b "
                                 "is in equilibrium but phase_equilibrium_form"
                                 " was not specified for all appropriate "
                                 "phase pairs."):
            # Also reverse order of phases for component a - this should pass
            # and component b should be flagged as missing
            m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"phase_equilibrium_form": {("p2", "p1"): "foo"}},
                    "b": {"phase_equilibrium_form": {(1, 2): "foo"}},
                    "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "phases_in_equilibrium": [("p1", "p2")],
                "phase_equilibrium_state": {("p1", "p2"): "whoop"},
                "base_units": base_units})

    @pytest.mark.unit
    def test_phases_in_equilibrium_no_formulation(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params Generic Property Package provided "
                           "with a phases_in_equilibrium argument but no "
                           "method was specified for "
                           "phase_equilibrium_state."):
            m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                    "b": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                    "c": {"phase_equilibrium_form": {("p1", "p2"): "foo"}}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "phases_in_equilibrium": [("p1", "p2")],
                "base_units": base_units})

    @pytest.mark.unit
    def test_phases_in_equilibrium_missing_pair_formulation(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params Generic Property Package provided "
                           "with a phases_in_equilibrium argument but "
                           "phase_equilibrium_state was not specified "
                           "for all phase pairs."):
            # Also reverse order of phases for component a - this should pass
            # and component b should be flagged as missing
            m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                    "b": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                    "c": {"phase_equilibrium_form": {("p1", "p2"): "foo"}}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"}},
                "state_definition": "baz",
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "phases_in_equilibrium": [("p1", "p2")],
                "phase_equilibrium_state": {(1, 2): "whoop"},
                "base_units": base_units})

    @pytest.mark.unit
    def test_parameter_construction_no_value(self):
        m = ConcreteModel()

        class test_class():
            # Mook up property method class for testing
            def build_parameters(c):
                c.test_var = Var()

        with pytest.raises(ConfigurationError,
                           match="params parameter test_var was not assigned "
                           "a value. Please check your configuration "
                           "arguments."):
            m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"dens_mol_liq_comp": test_class},
                    "b": {},
                    "c": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

    @pytest.mark.unit
    def test_parameter_construction_no_data(self):
        m = ConcreteModel()

        class test_class():
            # Mook up property method class for testing
            def build_parameters(c):
                c.config.parameter_data["test"]

        with pytest.raises(ConfigurationError,
                           match="params values were not defined for "
                           "parameter dens_mol_liq_comp in component a. "
                           "Please check the parameter_data argument to "
                           "ensure values are provided."):
            m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"dens_mol_liq_comp": test_class},
                    "b": {},
                    "c": {}},
                "phases": {
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

    @pytest.mark.unit
    def test_no_elements(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

        assert not hasattr(m.params, "element_list")
        assert not hasattr(m.params, "element_comp")

    @pytest.mark.unit
    def test_partial_elements(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params not all Components declared an "
                           "elemental_composition argument. Either all "
                           "Components must declare this, or none."):
            m.params = DummyParameterBlock(default={
                    "components": {"a": {"elemental_composition": {"e1": 1}},
                                   "b": {},
                                   "c": {}},
                    "phases": {
                        "p1": {"type": LiquidPhase,
                               "component_list": ["a", "b"],
                               "equation_of_state": DummyEoS},
                        "p2": {"equation_of_state": DummyEoS}},
                    "state_definition": modules[__name__],
                    "pressure_ref": 1e5,
                    "temperature_ref": 300,
                    "base_units": base_units})

    @pytest.mark.unit
    def test_elements_not_float(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params values in elemental_composition must "
                           "be integers \(not floats\)\: e1\: 2.0."):
            m.params = DummyParameterBlock(default={
                    "components": {"a": {"elemental_composition": {"e1": 2.0}},
                                   "b": {},
                                   "c": {}},
                    "phases": {
                        "p1": {"type": LiquidPhase,
                               "component_list": ["a", "b"],
                               "equation_of_state": DummyEoS},
                        "p2": {"equation_of_state": DummyEoS}},
                    "state_definition": modules[__name__],
                    "pressure_ref": 1e5,
                    "temperature_ref": 300,
                    "base_units": base_units})

    @pytest.mark.unit
    def test_elements(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"elemental_composition": {"e1": 1, "e2": 2}},
                    "b": {"elemental_composition": {"e3": 3, "e4": 4}},
                    "c": {"elemental_composition": {"e1": 5, "e3": 6}}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

        assert isinstance(m.params.element_list, Set)
        assert len(m.params.element_list) == 4
        assert m.params.element_comp == {
            "a": {"e1": 1, "e2": 2, "e3": 0, "e4": 0},
            "b": {"e1": 0, "e2": 0, "e3": 3, "e4": 4},
            "c": {"e1": 5, "e2": 0, "e3": 6, "e4": 0}}

    @pytest.mark.unit
    def test_henry(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"henry_component": {"p1": modules[__name__]}},
                    "b": {},
                    "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

        assert isinstance(m.params.a.test_param_p1, Var)
        assert m.params.a.test_param_p1.value == 42

    @pytest.mark.unit
    def test_henry_invalid_phase_name(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params component a was marked as a Henry's "
                           "Law component in phase p3, but this is not a "
                           "valid phase name."):
            m.params = DummyParameterBlock(default={
                    "components": {
                        "a": {"henry_component": {"p3": modules[__name__]}},
                        "b": {},
                        "c": {}},
                    "phases": {
                        "p1": {"type": LiquidPhase,
                               "component_list": ["a", "b"],
                               "equation_of_state": DummyEoS},
                        "p2": {"equation_of_state": DummyEoS}},
                    "state_definition": modules[__name__],
                    "pressure_ref": 1e5,
                    "temperature_ref": 300,
                    "base_units": base_units})

    @pytest.mark.unit
    def test_henry_invalid_phase_type(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params component a was marked as a Henry's "
                           "Law component in phase p2, but this is not a "
                           "Liquid phase."):
            m.params = DummyParameterBlock(default={
                    "components": {
                        "a": {"henry_component": {"p2": modules[__name__]}},
                        "b": {},
                        "c": {}},
                    "phases": {
                        "p1": {"type": LiquidPhase,
                               "component_list": ["a", "b"],
                               "equation_of_state": DummyEoS},
                        "p2": {"equation_of_state": DummyEoS}},
                    "state_definition": modules[__name__],
                    "pressure_ref": 1e5,
                    "temperature_ref": 300,
                    "base_units": base_units})

    @pytest.mark.unit
    def test_inherent_reactions(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units,
                "inherent_reactions": {
                    "e1": {"stoichiometry": {("p1", "a"): -3,
                                             ("p1", "b"): 4},
                           "heat_of_reaction": "foo",
                           "equilibrium_form": "foo"}}})

        assert m.params.has_inherent_reactions

        assert isinstance(m.params.inherent_reaction_idx, Set)
        assert len(m.params.inherent_reaction_idx) == 1
        assert m.params.inherent_reaction_idx == ["e1"]

        assert hasattr(m.params, "inherent_reaction_stoichiometry")
        assert len(m.params.inherent_reaction_stoichiometry) == 5
        assert m.params.inherent_reaction_stoichiometry == {
            ("e1", "p1", "a"): -3, ("e1", "p1", "b"): 4,
            ("e1", "p2", "a"): 0, ("e1", "p2", "b"): 0, ("e1", "p2", "c"): 0}

        assert isinstance(m.params.reaction_e1, Block)
        assert isinstance(m.params.reaction_e1.reaction_order, Var)
        assert len(m.params.reaction_e1.reaction_order) == 5
        for i in m.params.reaction_e1.reaction_order:
            order = {("p1", "a"): -3, ("p1", "b"): 4,
                     ("p2", "a"): 0, ("p2", "b"): 0, ("p2", "c"): 0}
            assert m.params.reaction_e1.reaction_order[i].value == order[i]

    @pytest.mark.unit
    def test_inherent_reactions_no_stoichiometry(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params inherent reaction e1 was not "
                           "provided with a stoichiometry configuration "
                           "argument."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units,
                "inherent_reactions": {
                    "e1": {"heat_of_reaction": "foo",
                           "equilibrium_form": "foo"}}})

    @pytest.mark.unit
    def test_inherent_reactions_invalid_phase_stoichiometry(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params stoichiometry for inherent "
                           "reaction e1 included unrecognised phase p7."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units,
                "inherent_reactions": {
                    "e1": {"stoichiometry": {("p7", "a"): -3,
                                             ("p1", "b"): 4},
                           "heat_of_reaction": "foo",
                           "equilibrium_form": "foo"}}})

    @pytest.mark.unit
    def test_inherent_reactions_invalid_component_stoichiometry(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params stoichiometry for inherent "
                           "reaction e1 included unrecognised component c7."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units,
                "inherent_reactions": {
                    "e1": {"stoichiometry": {("p1", "c7"): -3,
                                             ("p1", "b"): 4},
                           "heat_of_reaction": "foo",
                           "equilibrium_form": "foo"}}})

    @pytest.mark.unit
    def test_inherent_reactions_no_form(self):
        m = ConcreteModel()

        with pytest.raises(ConfigurationError,
                           match="params inherent reaction e1 was not "
                           "provided with a equilibrium_form configuration "
                           "argument."):
            m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units,
                "inherent_reactions": {
                    "e1": {"stoichiometry": {("p1", "a"): -3,
                                             ("p1", "b"): 4}}}})

    @pytest.mark.unit
    def test_default_scaling(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

        dsf = m.params.default_scaling_factor

        assert dsf[("temperature", None)] == 1e-2
        assert dsf[("pressure", None)] == 1e-5
        assert dsf[("dens_mol_phase", None)] == 1e-2
        assert dsf[("enth_mol", None)] == 1e-4
        assert dsf[("entr_mol", None)] == 1e-2
        assert dsf[("fug_phase_comp", None)] == 1e-4
        assert dsf[("fug_coeff_phase_comp", None)] == 1
        assert dsf[("gibbs_mol", None)] == 1e-4
        assert dsf[("mole_frac_comp", None)] == 1e3
        assert dsf[("mole_frac_phase_comp", None)] == 1e3
        assert dsf[("mw", None)] == 1e3
        assert dsf[("mw_phase", None)] == 1e3

    @pytest.mark.unit
    def test_default_scaling_convert(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"type": LiquidPhase,
                           "component_list": ["a", "b"],
                           "equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"time": pyunits.minutes,
                               "length": pyunits.foot,
                               "mass": pyunits.pound,
                               "amount": pyunits.mol,
                               "temperature": pyunits.degR}})

        dsf = m.params.default_scaling_factor

        assert dsf[("temperature", None)] == 1e-2
        assert dsf[("pressure", None)] == 1e-8
        assert dsf[("dens_mol_phase", None)] == 1
        assert dsf[("enth_mol", None)] == 1e-9
        assert dsf[("entr_mol", None)] == 1e-7
        assert dsf[("fug_phase_comp", None)] == 1e-7
        assert dsf[("fug_coeff_phase_comp", None)] == 1
        assert dsf[("gibbs_mol", None)] == 1e-9
        assert dsf[("mole_frac_comp", None)] == 1e3
        assert dsf[("mole_frac_phase_comp", None)] == 1e3
        assert dsf[("mw", None)] == 1e3
        assert dsf[("mw_phase", None)] == 1e3


# -----------------------------------------------------------------------------
# Dummy methods for testing build calls to sub-modules
def define_state(b):
    b.state_defined = True


class TestGenericStateBlock(object):
    @pytest.fixture()
    def frame(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "p1": {"equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": base_units})

        m.props = m.params.build_state_block([1],
                                             default={"defined_state": False})

        # Add necessary variables to state block
        m.props[1].flow_mol = Var(bounds=(0, 3000))
        m.props[1].pressure = Var(bounds=(1000, 3000))
        m.props[1].temperature = Var(bounds=(100, 200))
        m.props[1].mole_frac_phase_comp = Var(m.params.phase_list,
                                              m.params.component_list)
        m.props[1].phase_frac = Var(m.params.phase_list)

        return m

    @pytest.mark.unit
    def test_build(self, frame):
        assert isinstance(frame.props, Block)
        assert len(frame.props) == 1

        # Check for expected behaviour for dummy methods
        assert frame.props[1].state_defined
        assert isinstance(frame.props[1].dummy_var, Var)
        assert frame.props[1].eos_common == 2

    @pytest.mark.unit
    def test_basic_scaling(self, frame):
        frame.props[1].calculate_scaling_factors()

        assert frame.props[1].scaling_check

        assert len(frame.props[1].scaling_factor) == 8
        assert frame.props[1].scaling_factor[
            frame.props[1].temperature] == 0.01
        assert frame.props[1].scaling_factor[
            frame.props[1].pressure] == 1e-5
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["p1", "a"]] == 1000
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["p1", "b"]] == 1000
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["p1", "c"]] == 1000
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["p2", "a"]] == 1000
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["p2", "b"]] == 1000
        assert frame.props[1].scaling_factor[
            frame.props[1].mole_frac_phase_comp["p2", "c"]] == 1000

    @pytest.mark.unit
    def test_flows(self, frame):

        # Need to set the material flow basis for this test
        def set_flow_basis():
            return MaterialFlowBasis.molar
        frame.props[1].get_material_flow_basis = set_flow_basis

        frame.props[1].flow_mol.fix(100)
        frame.props[1].phase_frac['p1'].fix(0.75)
        frame.props[1].phase_frac['p2'].fix(0.25)
        frame.props[1].mole_frac_phase_comp['p1', 'a'].fix(0.1)
        frame.props[1].mole_frac_phase_comp['p1', 'b'].fix(0.2)
        frame.props[1].mole_frac_phase_comp['p1', 'c'].fix(0.7)
        frame.props[1].mole_frac_phase_comp['p2', 'c'].fix(0.1)
        frame.props[1].mole_frac_phase_comp['p2', 'b'].fix(0.2)
        frame.props[1].mole_frac_phase_comp['p2', 'a'].fix(0.7)
        frame.props[1].params.get_component('a').mw = 2
        frame.props[1].params.get_component('b').mw = 3
        frame.props[1].params.get_component('c').mw = 4

        assert value(frame.props[1].flow_vol) == pytest.approx(
            100/55e3, rel=1e-4)
        assert value(frame.props[1].flow_vol_phase['p1']) == \
            pytest.approx(100/55e3*0.75, rel=1e-4)
        assert value(frame.props[1].flow_vol_phase['p2']) == \
            pytest.approx(100/55e3*0.25, rel=1e-4)
        assert value(frame.props[1].flow_mol_comp['a']) == \
            pytest.approx(100*0.1*0.75 + 100*0.7*0.25, rel=1e-4)
        assert value(frame.props[1].flow_mol_comp['b']) == \
            pytest.approx(100*0.2*0.75 + 100*0.2*0.25, rel=1e-4)
        assert value(frame.props[1].flow_mol_comp['c']) == \
            pytest.approx(100*0.1*0.25 + 100*0.7*0.75, rel=1e-4)
        assert value(frame.props[1].mw_phase['p1']) == \
            pytest.approx(2*0.1 + 3*0.2 + 4*0.7, rel=1e-4)
        assert value(frame.props[1].mw) == \
            pytest.approx(0.75*(2*0.1 + 3*0.2 + 4*0.7) +
                          0.25*(2*0.7 + 3*0.2 + 4*0.1), rel=1e-4)
        assert value(frame.props[1].flow_mass) == \
            pytest.approx(100*0.75*(2*0.1 + 3*0.2 + 4*0.7) +
                          100*0.25*(2*0.7 + 3*0.2 + 4*0.1), rel=1e-4)
        assert value(frame.props[1].flow_mass_phase["p1"]) == \
            pytest.approx(100*0.75*(2*0.1 + 3*0.2 + 4*0.7), rel=1e-4)
        assert value(frame.props[1].flow_mass_phase["p2"]) == \
            pytest.approx(100*0.25*(2*0.7 + 3*0.2 + 4*0.1), rel=1e-4)

    class dummy_prop():
        def return_expression(*args, **kwargs):
            return 4

    @pytest.mark.unit
    def test_diffus_phase_comp(self, frame):
        frame.params.p1.config.diffus_phase_comp = \
            TestGenericStateBlock.dummy_prop
        frame.params.p2.config.diffus_phase_comp = \
            TestGenericStateBlock.dummy_prop

        for p, j in frame.props[1].phase_component_set:
            assert value(frame.props[1].diffus_phase_comp[p, j]) == 4

    @pytest.mark.unit
    def test_visc_d_phase(self, frame):
        frame.params.p1.config.visc_d_phase = \
            TestGenericStateBlock.dummy_prop
        frame.params.p2.config.visc_d_phase = \
            TestGenericStateBlock.dummy_prop

        for p in frame.props[1].phase_list:
            assert value(frame.props[1].visc_d_phase[p]) == 4



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Implementation of the formulation proposed in:

Burgard, A.P., Eason, J.P., Eslick, J.C., Ghouse, J.H., Lee, A., Biegler, L.T.,
Miller, D.C., 2018, A Smooth, Square Flash Formulation for Equation-Oriented
Flowsheet Optimization. Proceedings of the 13th International Symposium on
Process Systems Engineering – PSE 2018, July 1-5, 2018, San Diego.
"""
from pyomo.environ import Constraint, Param, Var, value
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.math import smooth_max, smooth_min
from idaes.generic_models.properties.core.phase_equil.bubble_dew import (
    _valid_VL_component_list)
import idaes.core.util.scaling as iscale


def phase_equil(b, phase_pair):
    # This method is called via StateBlock.build, thus does not need clean-up
    # try/except statements
    suffix = "_"+phase_pair[0]+"_"+phase_pair[1]

    # Smooth VLE assumes a liquid and a vapor phase, so validate this
    (l_phase,
     v_phase,
     vl_comps,
     henry_comps,
     l_only_comps,
     v_only_comps) = _valid_VL_component_list(b, phase_pair)

    if l_phase is None or v_phase is None:
        raise ConfigurationError(
            "{} Generic Property Package phase pair {}-{} was set to use "
            "Smooth VLE formulation, however this is not a vapor-liquid pair."
            .format(b.params.name, phase_pair[0], phase_pair[1]))

    # Definition of equilibrium temperature for smooth VLE
    t_units = b.params.get_metadata().default_units["temperature"]
    if v_only_comps == []:
        b.add_component("_t1"+suffix, Var(
                initialize=b.temperature.value,
                doc='Intermediate temperature for calculating Teq',
                units=t_units))
        _t1 = getattr(b, "_t1"+suffix)

        b.add_component("eps_1"+suffix, Param(
            default=0.01,
            mutable=True,
            doc='Smoothing parameter for Teq',
            units=t_units))
        eps_1 = getattr(b, "eps_1"+suffix)

        # PSE paper Eqn 13
        def rule_t1(b):
            return _t1 == smooth_max(b.temperature,
                                     b.temperature_bubble[phase_pair],
                                     eps_1)
        b.add_component("_t1_constraint"+suffix, Constraint(rule=rule_t1))
    else:
        _t1 = b.temperature

    if l_only_comps == []:
        b.add_component("eps_2"+suffix, Param(
            default=0.0005, mutable=True, doc='Smoothing parameter for Teq',
            units=t_units))
        eps_2 = getattr(b, "eps_2"+suffix)

        # PSE paper Eqn 14
        # TODO : Add option for supercritical extension
        def rule_teq(b):
            return b._teq[phase_pair] == smooth_min(
                _t1, b.temperature_dew[phase_pair], eps_2)
    elif v_only_comps == []:
        def rule_teq(b):
            return b._teq[phase_pair] == _t1
    else:
        def rule_teq(b):
            return b._teq[phase_pair] == b.temperature
    b.add_component("_teq_constraint"+suffix, Constraint(rule=rule_teq))


def calculate_scaling_factors(b, phase_pair):
    suffix = "_"+phase_pair[0]+"_"+phase_pair[1]
    sf_T = iscale.get_scaling_factor(
            b.temperature, default=1, warning=True)

    try:
        _t1 = getattr(b, "_t1"+suffix)
        _t1_cons = getattr(b, "_t1_constraint"+suffix)
        iscale.set_scaling_factor(_t1, sf_T)
        iscale.constraint_scaling_transform(_t1_cons, sf_T, overwrite=False)
    except AttributeError:
        pass

    _teq_cons = getattr(b, "_teq_constraint"+suffix)
    iscale.constraint_scaling_transform(_teq_cons, sf_T, overwrite=False)


def phase_equil_initialization(b, phase_pair):
    suffix = "_"+phase_pair[0]+"_"+phase_pair[1]

    for c in b.component_objects(Constraint):
        # Activate equilibrium constraints
        if c.local_name in ("_t1_constraint"+suffix,
                            "_teq_constraint"+suffix):
            c.activate()


def calculate_teq(b, phase_pair):
    suffix = "_"+phase_pair[0]+"_"+phase_pair[1]

    if hasattr(b, "eq_temperature_bubble"):
        _t1 = getattr(b, "_t1"+suffix)
        _t1.value = max(value(b.temperature),
                        b.temperature_bubble[phase_pair].value)
    else:
        _t1 = b.temperature

    if hasattr(b, "eq_temperature_dew"):
        b._teq[phase_pair].value = min(_t1.value,
                                       b.temperature_dew[phase_pair].value)
    else:
        b._teq[phase_pair].value = _t1.value


# -----------------------------------------------------------------------------
class SmoothVLE(object):
    # Deprecation: Eventually replace static methods with class methods
    phase_equil = phase_equil
    phase_equil_initialization = phase_equil_initialization
    calculate_teq = calculate_teq
    calculate_scaling_factors = calculate_scaling_factors



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

__author__ = "Costing Team (A. Noring and M. Zamarripa)"
__version__ = "1.0.0"

import pytest

from idaes.power_generation.costing.power_plant_costing import \
    (get_sCO2_unit_cost,
     get_PP_costing,
     get_ASU_cost,
     build_flowsheet_cost_constraint,
     costing_initialization)
from idaes.core.util.model_statistics import (degrees_of_freedom)
import pyomo.environ as pyo
from idaes.generic_models.properties import iapws95
from idaes.generic_models.properties import swco2
from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver


@pytest.mark.component
def test_PP_costing():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.get_costing(year='2018')

    # check that the model solved properly and has 0 degrees of freedom
    assert(degrees_of_freedom(m) == 0)

    ###########################################################################
    #  Create costing constraints                                             #
    ###########################################################################

    # coal flow rate
    # accounts 1.x and 2.x are coal handling, preparation and feed
    # accounts 4.x are for boiler BOP and foundations
    coal_accounts = ['1.1', '1.2', '1.3', '1.4',
                     '2.1', '2.2', '4.11', '4.15', '4.16']
    m.fs.boiler = pyo.Block()
    m.fs.boiler.coal_mass_flow = pyo.Var(initialize=7238.95)  # tpd
    m.fs.boiler.coal_mass_flow.fix()
    get_PP_costing(m.fs.boiler, coal_accounts, m.fs.boiler.coal_mass_flow,
                   'tpd', 2)

    # total fuel feed
    # accounts 3.x are for start up systems and miscellaneous plant equipment
    # accounts 7.x are for ductwork and stack foundations
    fuel_accounts = ['3.6', '3.9', '7.3', '7.5']
    m.fs.fuel_feed = pyo.Block()
    m.fs.fuel_feed.total_fuel_feed = pyo.Var(initialize=603246)  # lb/hr
    m.fs.fuel_feed.total_fuel_feed.fix()
    get_PP_costing(m.fs.fuel_feed, fuel_accounts,
                   m.fs.fuel_feed.total_fuel_feed, 'lb/hr', 2)

    # HP BFW flow rate
    # accounts 3.x are for feedwater systems
    # account 4.9 is for the boiler
    # account 8.4 is steam piping
    BFW_accounts = ['3.1', '3.3', '3.5', '4.9', '8.4']
    m.fs.bfp = pyo.Block()
    m.fs.bfp.BFW_mass_flow = pyo.Var(initialize=5316158)  # lb/hr
    m.fs.bfp.BFW_mass_flow.fix()
    get_PP_costing(m.fs.bfp, BFW_accounts, m.fs.bfp.BFW_mass_flow, 'lb/hr', 2)

    # Steam turbine power
    # accounts 8.x are for the steam turbine and its foundations
    power_accounts = ['8.1']
    m.fs.turb = pyo.Block()
    m.fs.turb.power = pyo.Var(initialize=769600)  # kW
    m.fs.turb.power.fix()
    get_PP_costing(m.fs.turb, power_accounts, m.fs.turb.power, 'kW', 2)

    # Condernser duty
    cond_accounts = ['8.3']
    m.fs.condenser = pyo.Block()
    m.fs.condenser.duty_MMBtu = pyo.Var(initialize=2016)  # MMBtu/hr
    m.fs.condenser.duty_MMBtu.fix()
    get_PP_costing(m.fs.condenser, cond_accounts, m.fs.condenser.duty_MMBtu,
                   "MMBtu/hr", 2)

    # Circulating water flow rate
    # accounts 9.x are for circulating water systems
    # account 14.5 is for the pumphouse
    circ_accounts = ['9.2', '9.3', '9.4', '9.6', '9.7', '14.5']
    m.fs.circulating_water = pyo.Block()
    m.fs.circulating_water.vol_flow = pyo.Var(initialize=463371)  # gpm
    m.fs.circulating_water.vol_flow.fix()
    get_PP_costing(m.fs.circulating_water,
                   circ_accounts, m.fs.circulating_water.vol_flow, 'gpm', 2)

    # Ash flow rate
    # accounts are for ash storage and handling
    ash_accounts = ['10.6', '10.7', '10.9']
    m.fs.ash_handling = pyo.Block()
    m.fs.ash_handling.ash_mass_flow = pyo.Var(initialize=66903)  # lb/hr
    m.fs.ash_handling.ash_mass_flow.fix()
    get_PP_costing(m.fs.ash_handling, ash_accounts,
                   m.fs.ash_handling.ash_mass_flow, 'lb/hr', 2)

    # add total cost
    build_flowsheet_cost_constraint(m)

    # add initialize
    costing_initialization(m.fs)

    # try solving
    solver = get_solver()
    results = solver.solve(m, tee=True)

    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal

    #  all numbers come from the NETL excel file:
    # "201.001.001_BBR4 COE Spreadsheet_Rev0U_20190919_njk.xlsm"
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['1.1']), abs=1e-1) == 2306/1e3
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['1.2']), abs=1e-1) == 6385/1e3
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['1.3']), abs=1e-1) == 59527/1e3
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['1.4']), abs=1e-1) == 8086/1e3
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['2.1']), abs=1e-1) == 4073/1e3
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['2.2']), abs=1e-1) == 13976/1e3
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['4.11']), abs=1e-1) == 3751/1e3
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['4.15']), abs=1e-1) == 197/1e3
    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.total_plant_cost['4.16']), abs=1e-1) == 1014/1e3

    assert pytest.approx(pyo.value(
        m.fs.fuel_feed.costing.total_plant_cost['3.6']), abs=1e-1) == 4864/1e3
    assert pytest.approx(pyo.value(
        m.fs.fuel_feed.costing.total_plant_cost['3.9']), abs=1e-1) == 522/1e3
    assert pytest.approx(pyo.value(
        m.fs.fuel_feed.costing.total_plant_cost['7.3']), abs=1e-1) == 1710/1e3
    assert pytest.approx(pyo.value(
        m.fs.fuel_feed.costing.total_plant_cost['7.5']), abs=1e-1) == 647/1e3

    assert pytest.approx(pyo.value(
        m.fs.bfp.costing.total_plant_cost['3.1']), abs=1e-1) == 19233/1e3
    assert pytest.approx(pyo.value(
        m.fs.bfp.costing.total_plant_cost['3.3']), abs=1e-1) == 6897/1e3
    assert pytest.approx(pyo.value(
        m.fs.bfp.costing.total_plant_cost['3.5']), abs=1e-1) == 2366/1e3
    assert pytest.approx(pyo.value(
        m.fs.bfp.costing.total_plant_cost['4.9']), abs=1e-1) == 570418/1e3
    assert pytest.approx(pyo.value(
        m.fs.bfp.costing.total_plant_cost['8.4']), abs=1e-1) == 81916/1e3

    assert pytest.approx(pyo.value(
        m.fs.turb.costing.total_plant_cost['8.1']), abs=1e-1) == 110166/1e3

    assert pytest.approx(pyo.value(
        m.fs.condenser.costing.total_plant_cost['8.3']), abs=1e-1) == 20447/1e3

    assert pytest.approx(pyo.value(
        m.fs.circulating_water.costing.total_plant_cost['9.2']),
        abs=1e-1) == 4133/1e3
    assert pytest.approx(pyo.value(
        m.fs.circulating_water.costing.total_plant_cost['9.3']),
        abs=1e-1) == 25518/1e3
    assert pytest.approx(pyo.value(
        m.fs.circulating_water.costing.total_plant_cost['9.4']),
        abs=1e-1) == 19859/1e3
    assert pytest.approx(pyo.value(
        m.fs.circulating_water.costing.total_plant_cost['9.6']),
        abs=1e-1) == 2870/1e3
    assert pytest.approx(pyo.value(
        m.fs.circulating_water.costing.total_plant_cost['9.7']),
        abs=1e-1) == 2690/1e3
    assert pytest.approx(pyo.value(
        m.fs.circulating_water.costing.total_plant_cost['14.5']),
        abs=1e-1) == 464/1e3

    assert pytest.approx(pyo.value(
        m.fs.ash_handling.costing.total_plant_cost['10.6']),
        abs=1e-1) == 6429/1e3
    assert pytest.approx(pyo.value(
        m.fs.ash_handling.costing.total_plant_cost['10.7']),
        abs=1e-1) == 10725/1e3
    assert pytest.approx(pyo.value(
        m.fs.ash_handling.costing.total_plant_cost['10.9']),
        abs=1e-1) == 2564/1e3

    assert pytest.approx(pyo.value(
        m.fs.flowsheet_cost), abs=1e-1) == 993753/1e3

    return m


@pytest.mark.component
def test_power_plant_costing():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.get_costing(year='2018')

    ###########################################################################
    #  Create costing constraints                                             #
    ###########################################################################

    # subcritical PC
    coal_accounts = ['1.1', '1.2', '1.3']
    m.fs.subcritical_PC = pyo.Block()
    m.fs.subcritical_PC.coal_feed_rate = pyo.Var(initialize=7613.37)  # tpd
    m.fs.subcritical_PC.coal_feed_rate.fix()
    get_PP_costing(m.fs.subcritical_PC, coal_accounts,
                   m.fs.subcritical_PC.coal_feed_rate,
                   'tpd', 1)

    # two-stage, slurry-feed IGCC
    feedwater_accounts = ['3.1', '3.3', '3.5']
    m.fs.IGCC_1 = pyo.Block()
    m.fs.IGCC_1.feedwater_flow_rate = pyo.Var(initialize=1576062.15)  # lb/hr
    m.fs.IGCC_1.feedwater_flow_rate.fix()
    get_PP_costing(m.fs.IGCC_1, feedwater_accounts,
                   m.fs.IGCC_1.feedwater_flow_rate,
                   'lb/hr', 3)

    # single-stage, slurry-feed, IGCC
    syngas_accounts = ['6.1', '6.2', '6.3']
    m.fs.IGCC_2 = pyo.Block()
    m.fs.IGCC_2.syngas_flow_rate = pyo.Var(initialize=182335.921)  # lb/hr
    m.fs.IGCC_2.syngas_flow_rate.fix()
    get_PP_costing(m.fs.IGCC_2, syngas_accounts,
                   m.fs.IGCC_2.syngas_flow_rate,
                   'lb/hr', 4)

    # single-stage, dry-feed, IGCC
    HRSG_accounts = ['7.1', '7.2']
    m.fs.IGCC_3 = pyo.Block()
    m.fs.IGCC_3.HRSG_duty = pyo.Var(initialize=1777.86)  # MMBtu/hr
    m.fs.IGCC_3.HRSG_duty.fix()
    get_PP_costing(m.fs.IGCC_3, HRSG_accounts,
                   m.fs.IGCC_3.HRSG_duty,
                   'MMBtu/hr', 5)

    # NGCC
    steam_turbine_accounts = ['8.1', '8.2', '8.5']
    m.fs.NGCC = pyo.Block()
    m.fs.NGCC.turbine_power = pyo.Var(initialize=212500)  # kW
    m.fs.NGCC.turbine_power.fix()
    get_PP_costing(m.fs.NGCC, steam_turbine_accounts,
                   m.fs.NGCC.turbine_power,
                   'kW', 6)

    # AUSC PC
    AUSC_accounts = ['4.9', '8.4']
    m.fs.AUSC = pyo.Block()
    m.fs.AUSC.feedwater_flow = pyo.Var(initialize=3298815.58)  # lb/hr
    m.fs.AUSC.feedwater_flow.fix()
    get_PP_costing(m.fs.AUSC, AUSC_accounts,
                   m.fs.AUSC.feedwater_flow,
                   'lb/hr', 7)

    # add total cost
    build_flowsheet_cost_constraint(m)

    # add initialize
    costing_initialization(m.fs)

    # try solving
    solver = get_solver()
    results = solver.solve(m, tee=True)

    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    #  all numbers come from the NETL excel file
    # "201.001.001_BBR4 COE Spreadsheet_Rev0U_20190919_njk.xlsm"
    assert pytest.approx(pyo.value(
        m.fs.subcritical_PC.costing.total_plant_cost['1.1']),
        abs=1e-1) == 2379/1e3
    assert pytest.approx(pyo.value(
        m.fs.subcritical_PC.costing.total_plant_cost['1.2']),
        abs=1e-1) == 6588/1e3
    assert pytest.approx(pyo.value(
        m.fs.subcritical_PC.costing.total_plant_cost['1.3']),
        abs=1e-1) == 61409/1e3

    assert pytest.approx(pyo.value(
        m.fs.IGCC_1.costing.total_plant_cost['3.1']), abs=1e-1) == 10807/1e3
    assert pytest.approx(pyo.value(
        m.fs.IGCC_1.costing.total_plant_cost['3.3']), abs=1e-1) == 2564/1e3
    assert pytest.approx(pyo.value(
        m.fs.IGCC_1.costing.total_plant_cost['3.5']), abs=1e-1) == 923/1e3

    assert pytest.approx(pyo.value(
        m.fs.IGCC_2.costing.total_plant_cost['6.1']), abs=1e-1) == 117850/1e3
    assert pytest.approx(pyo.value(
        m.fs.IGCC_2.costing.total_plant_cost['6.2']), abs=1e-1) == 3207/1e3
    assert pytest.approx(pyo.value(
        m.fs.IGCC_2.costing.total_plant_cost['6.3']), abs=1e-1) == 3770/1e3

    assert pytest.approx(pyo.value(
        m.fs.IGCC_3.costing.total_plant_cost['7.1']), abs=1e-1) == 53530/1e3
    assert pytest.approx(pyo.value(
        m.fs.IGCC_3.costing.total_plant_cost['7.2']), abs=1e-1) == 19113/1e3

    assert pytest.approx(pyo.value(
        m.fs.NGCC.costing.total_plant_cost['8.1']), abs=1e-1) == 49468/1e3
    assert pytest.approx(pyo.value(
        m.fs.NGCC.costing.total_plant_cost['8.2']), abs=1e-1) == 565/1e3
    assert pytest.approx(pyo.value(
        m.fs.NGCC.costing.total_plant_cost['8.5']), abs=1e-1) == 4094/1e3

    assert pytest.approx(pyo.value(
        m.fs.AUSC.costing.bare_erected_cost['4.9']), abs=1e-1) == 295509/1e3
    assert pytest.approx(pyo.value(
        m.fs.AUSC.costing.bare_erected_cost['8.4']), abs=1e-1) == 57265/1e3

    return m


@pytest.mark.component
def test_sCO2_costing():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.get_costing(year='2017')

    # ######################################################
    # Primary Heater
    m.fs.boiler = pyo.Block()
    m.fs.boiler.heat_duty = pyo.Var(initialize=1461.5e6)
    m.fs.boiler.heat_duty.fix()
    m.fs.boiler.temp = pyo.Var(initialize=620)  # C
    m.fs.boiler.temp.fix()
    get_sCO2_unit_cost(m.fs.boiler, 'Coal-fired heater',
                       m.fs.boiler.heat_duty*(1e-6),
                       temp_C=m.fs.boiler.temp)

    # ######################################################
    # CO2 Turbine
    m.fs.turbine = pyo.Block()
    m.fs.turbine.work_isentropic = pyo.Var(initialize=1006.2e6)
    m.fs.turbine.work_isentropic.fix()
    m.fs.turbine.temp = pyo.Var(initialize=620)
    m.fs.turbine.temp.fix()
    get_sCO2_unit_cost(m.fs.turbine, 'Axial turbine',
                       m.fs.turbine.work_isentropic*(1e-6),
                       temp_C=m.fs.turbine.temp,
                       n_equip=1)

    # ######################################################
    # Generator
    m.fs.generator = pyo.Block()
    m.fs.generator.work_isentropic = pyo.Var(initialize=1006.2e6)
    get_sCO2_unit_cost(m.fs.generator, 'Generator',
                       m.fs.turbine.work_isentropic*(1e-6),
                       n_equip=1)

    # ######################################################
    # High Temperature Recuperator
    m.fs.HTR = pyo.Block()
    m.fs.HTR.heat_duty = pyo.Var(initialize=1461e6)  # W
    m.fs.HTR.heat_duty.fix()
    m.fs.HTR.LMTD = pyo.Var(initialize=21.45)
    m.fs.HTR.LMTD.fix()
    m.fs.HTR.temp = pyo.Var(initialize=453)
    m.fs.HTR.temp.fix()

    m.fs.HTR.UA = pyo.Var(initialize=1e8)

    # gives units of W/K
    @m.fs.Constraint()
    def HTR_UA_rule(b):
        return (b.HTR.UA*b.HTR.LMTD == b.HTR.heat_duty)

    get_sCO2_unit_cost(m.fs.HTR, 'Recuperator', m.fs.HTR.UA,
                       temp_C=m.fs.HTR.temp)

    # ######################################################
    # Low Temperature Recuperator
    m.fs.LTR = pyo.Block()
    m.fs.LTR.heat_duty = pyo.Var(initialize=911.7e6)  # W
    m.fs.LTR.heat_duty.fix()
    m.fs.LTR.LMTD = pyo.Var(initialize=5.21)
    m.fs.LTR.LMTD.fix()
    m.fs.LTR.temp = pyo.Var(initialize=216)
    m.fs.LTR.temp.fix()
    m.fs.LTR.UA = pyo.Var(initialize=1e8)

    @m.fs.Constraint()
    def LTR_UA_rule(b):
        return (b.LTR.UA*b.LTR.LMTD == b.LTR.heat_duty)

    get_sCO2_unit_cost(m.fs.LTR, 'Recuperator', m.fs.LTR.UA,
                       temp_C=m.fs.LTR.temp)

    # ######################################################
    # CO2 Cooler, costed using the recouperator not dry cooler
    m.fs.co2_cooler = pyo.Block()
    m.fs.co2_cooler.heat_duty = pyo.Var(initialize=739421217)
    m.fs.co2_cooler.heat_duty.fix()
    m.fs.co2_cooler.temp = pyo.Var(initialize=81)
    m.fs.co2_cooler.temp.fix()

    # Estimating LMTD
    # Cost from report: $27,780 thousand
    # Back-calculated UA: 41819213 W/K
    # Heat duty from report: 2523 MMBTu/hr --> 739421217 W
    # Estimated LMTD: 17.68 K
    m.fs.co2_cooler.LMTD = pyo.Var(initialize=5)
    m.fs.co2_cooler.UA = pyo.Var(initialize=1e5)
    m.fs.co2_cooler.LMTD.fix(17.68)

    @m.fs.Constraint()
    def co2_cooler_UA_rule(b):
        return (b.co2_cooler.UA * b.co2_cooler.LMTD ==
                b.co2_cooler.heat_duty)

    get_sCO2_unit_cost(m.fs.co2_cooler, 'Recuperator',
                       m.fs.co2_cooler.UA,
                       temp_C=m.fs.co2_cooler.temp)

    # ######################################################
    # Main Compressor - 5.99 m^3/s in Baseline620
    m.fs.main_compressor = pyo.Block()
    m.fs.main_compressor.flow_vol = pyo.Var(initialize=5.99)
    m.fs.main_compressor.flow_vol.fix()
    get_sCO2_unit_cost(m.fs.main_compressor, 'Barrel type compressor',
                       m.fs.main_compressor.flow_vol,
                       n_equip=5.0)

    # ######################################################
    # Main Compressor Motor
    m.fs.main_compressor_motor = pyo.Block()
    m.fs.main_compressor_motor.work_isentropic = pyo.Var(initialize=159.7e6)
    m.fs.main_compressor_motor.work_isentropic.fix()
    get_sCO2_unit_cost(m.fs.main_compressor_motor, 'Open drip-proof motor',
                       m.fs.main_compressor_motor.work_isentropic*(1e-6),
                       n_equip=5.0)

    # ######################################################
    # Recompressor - 6.89 m^3/s in Baseline620
    m.fs.bypass_compressor = pyo.Block()
    m.fs.bypass_compressor.flow_vol = pyo.Var(initialize=6.89)
    m.fs.bypass_compressor.flow_vol.fix()
    get_sCO2_unit_cost(m.fs.bypass_compressor, 'Barrel type compressor',
                       m.fs.bypass_compressor.flow_vol,
                       n_equip=4.0)

    # ######################################################
    # Recompressor Motor
    m.fs.bypass_compressor_motor = pyo.Block()
    m.fs.bypass_compressor_motor.work_isentropic = pyo.Var(initialize=124.3e6)
    m.fs.bypass_compressor_motor.work_isentropic.fix()

    get_sCO2_unit_cost(m.fs.bypass_compressor_motor, 'Open drip-proof motor',
                       m.fs.bypass_compressor_motor.work_isentropic*(1e-6),
                       n_equip=4.0)

    # add total cost
    build_flowsheet_cost_constraint(m)

    # add initialize
    costing_initialization(m.fs)

    # try solving
    solver = get_solver()
    results = solver.solve(m, tee=True)

    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal

    assert pytest.approx(pyo.value(
        m.fs.boiler.costing.equipment_cost), abs=1e-1) == 216300/1e3
    assert pytest.approx(pyo.value(
        m.fs.turbine.costing.equipment_cost), abs=1e-1) == 13160/1e3
    assert pytest.approx(pyo.value(
        m.fs.generator.costing.equipment_cost), abs=1e-1) == 4756/1e3
    assert pytest.approx(pyo.value(
        m.fs.HTR.costing.equipment_cost), abs=1e-1) == 40150/1e3
    assert pytest.approx(pyo.value(
        m.fs.LTR.costing.equipment_cost), abs=1e-1) == 81860/1e3
    assert pytest.approx(pyo.value(
        m.fs.co2_cooler.costing.equipment_cost), abs=1e-1) == 27780/1e3
    assert pytest.approx(pyo.value(
        m.fs.main_compressor.costing.equipment_cost), abs=1e-1) == 31640/1e3
    assert pytest.approx(pyo.value(
        m.fs.bypass_compressor.costing.equipment_cost), abs=1e-1) == 26360/1e3
    assert pytest.approx(
        pyo.value(m.fs.main_compressor_motor.costing.equipment_cost) +
        pyo.value(m.fs.bypass_compressor_motor.costing.equipment_cost),
        abs=1e-1) == 29130/1e3
    assert pytest.approx(pyo.value(
        m.fs.bypass_compressor.costing.equipment_cost), abs=1e-1) == 26360/1e3

    return m


@pytest.mark.component
def test_ASU_costing():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.get_costing(year='2017')

    m.fs.ASU = pyo.Block()
    m.fs.ASU.O2_flow = pyo.Var()
    m.fs.ASU.O2_flow.fix(13078)  # TPD

    get_ASU_cost(m.fs.ASU, m.fs.ASU.O2_flow)

    # try solving
    solver = get_solver()
    results = solver.solve(m, tee=True)

    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal

    m.fs.ASU.costing.bare_erected_cost.display()

    assert pytest.approx(pyo.value(
        m.fs.ASU.costing.bare_erected_cost), abs=1) == 3.2675e6/1e3

    return m
